/*
 * SPDX-License-Identifier: Apache-2.0
 */

#define DT_DRV_COMPAT juju_spi_bitbang_direct

#include "zephyr/drivers/gpio.h"
#include <ch32v00x.h>
#include <zephyr/drivers/spi.h>

struct spi_bitbang_data {
	struct gpio_dt_spec cs;
};

struct spi_bitbang_config {
	void *regs;
	const struct gpio_dt_spec clk;
	const struct gpio_dt_spec mosi;
	const struct gpio_dt_spec *cs;
	uint8_t num_cs;
};

static void spi_bitbang_write_bits(void *regs, uint32_t mask, uint32_t bits)
{
	((volatile GPIO_TypeDef *)regs)->BSHR = ((~bits & mask) << 16) | bits;
}

static __attribute__((optimize("-O2"))) void
spi_bitbang_write_word(void *regs, uint32_t w, int bits, uint32_t mask,
		       uint32_t mosi_high, uint32_t mosi_low,
		       uint32_t clock_low, uint32_t clock_high)
{
	for (; bits != 0;) {
		uint32_t mosi =
			((w & 0x80000000) != 0) ? mosi_high : mosi_low;
		spi_bitbang_write_bits(regs, mask, mosi | clock_low);
		w <<= 1;
		bits--;
		spi_bitbang_write_bits(regs, mask, mosi | clock_high);
	}
}

static __attribute__((optimize("-O2"))) void
spi_bitbang_write_buf(void *regs, const struct spi_buf *tx_buf, uint32_t mask,
		      uint32_t mosi_high, uint32_t mosi_low,
		      uint32_t clock_low, uint32_t clock_high)
{
	const uint8_t *p = tx_buf->buf;
	size_t j;
	int bits = 0;
	uint32_t w = 0;

	for (j = 0; j < tx_buf->len; j++) {
		if (bits == 32) {
			spi_bitbang_write_word(regs, w, bits, mask, mosi_high,
					       mosi_low, clock_low,
					       clock_high);
			bits = 0;
			w = 0;
		}
		w <<= 8;
		w |= *p++;
		bits += 8;
	}
	if (bits != 0) {
		spi_bitbang_write_word(regs, w << (32 - bits), bits, mask,
				       mosi_high, mosi_low, clock_low,
				       clock_high);
	}
}

static int spi_bitbang_transmit(const struct device *dev,
				const struct spi_config *spi_config,
				const struct spi_buf_set *tx_bufs)
{
	const struct spi_bitbang_config *info = dev->config;
	const struct gpio_dt_spec *mosi = &info->mosi;
	const struct device *port = mosi->port;
	struct spi_bitbang_data *data = dev->data;
	gpio_port_pins_t clock_bit = (1 << info->clk.pin);
	gpio_port_pins_t mosi_bit = (1 << mosi->pin);
	gpio_port_pins_t mask = clock_bit | mosi_bit;
	gpio_port_pins_t invert =
		((const struct gpio_driver_data *)port->data)->invert & mask;
	gpio_port_pins_t mosi_low = (invert & mosi_bit);
	gpio_port_pins_t mosi_high = mosi_low ^ mosi_bit;
	gpio_port_pins_t clock_low = (invert & mosi_bit);
	gpio_port_pins_t clock_high = clock_low ^ clock_bit;
	bool cpol =
		(SPI_MODE_GET(spi_config->operation) & SPI_MODE_CPOL) != 0;
	bool cpha =
		(SPI_MODE_GET(spi_config->operation) & SPI_MODE_CPHA) != 0;
	size_t i;

	if (cpha != cpol) {
		clock_low ^= clock_bit;
		clock_high ^= clock_bit;
	}

	gpio_port_set_masked_raw(port, mask, cpha ? clock_high : clock_low);

	data->cs = spi_config->cs.gpio;
	if (data->cs.port != NULL) {
		gpio_pin_configure_dt(&data->cs, GPIO_OUTPUT_ACTIVE);
	}

	if (cpha) {
		gpio_port_set_masked_raw(port, mask, clock_low);
	}

	for (i = 0; i < tx_bufs->count; i++) {
		spi_bitbang_write_buf(info->regs, &tx_bufs->buffers[i], mask,
				      mosi_high, mosi_low, clock_low,
				      clock_high);
	}

	if (data->cs.port != NULL &&
	    (spi_config->operation & SPI_HOLD_ON_CS) == 0) {
		gpio_pin_set_dt(&data->cs, false);
	}

	return 0;
}

static int spi_bitbang_transceive(const struct device *dev,
				  const struct spi_config *config,
				  const struct spi_buf_set *tx_bufs,
				  const struct spi_buf_set *rx_bufs)
{
	if (config->operation & SPI_OP_MODE_SLAVE) {
		return -ENOTSUP;
	}

	if ((config->operation & (SPI_MODE_LOOP | SPI_HALF_DUPLEX)) != 0) {
		return -ENOTSUP;
	}

	if ((config->operation &
	     (SPI_TRANSFER_LSB | SPI_LINES_DUAL | SPI_LINES_QUAD))) {
		return -ENOTSUP;
	}

	if (SPI_WORD_SIZE_GET(config->operation) != 8) {
		return -ENOTSUP;
	}

	if (config->frequency != 0 && config->frequency < 1000000) {
		return -ENOTSUP;
	}

	if (rx_bufs != NULL) {
		return -ENOTSUP;
	}

	return spi_bitbang_transmit(dev, config, tx_bufs);
}

static int spi_bitbang_release(const struct device *dev,
			       const struct spi_config *config)
{
	struct spi_bitbang_data *data = dev->data;

	if (data->cs.port != NULL) {
		gpio_pin_set_dt(&data->cs, false);
	}

	return 0;
}

static const struct spi_driver_api spi_bitbang_api = {
	.transceive = spi_bitbang_transceive,
	.release = spi_bitbang_release,
};

static int spi_bitbang_init(const struct device *dev)
{
	const struct spi_bitbang_config *config = dev->config;
	int i;
	int err;

	if (!gpio_is_ready_dt(&config->clk)) {
		return -ENODEV;
	}
	err = gpio_pin_configure_dt(&config->clk, GPIO_OUTPUT_INACTIVE);
	if (err < 0) {
		return err;
	}

	if (!gpio_is_ready_dt(&config->mosi)) {
		return -ENODEV;
	}
	err = gpio_pin_configure_dt(&config->mosi, GPIO_OUTPUT_INACTIVE);
	if (err < 0) {
		return err;
	}

	for (i = 0; i < config->num_cs; i++) {
		const struct gpio_dt_spec *cs = &config->cs[i];
		if (!gpio_is_ready_dt(cs)) {
			return -ENODEV;
		}

		err = gpio_pin_configure_dt(cs, GPIO_OUTPUT_INACTIVE);
		if (err < 0) {
			return err;
		}
	}

	return 0;
}

#define SPI_CONTEXT_CS_GPIO_SPEC_ELEM(_node_id, _prop, _idx)                 \
	GPIO_DT_SPEC_GET_BY_IDX(_node_id, _prop, _idx),

#define SPI_CONTEXT_CS_GPIOS_FOREACH_ELEM(_node_id)                          \
	DT_FOREACH_PROP_ELEM(_node_id, cs_gpios,                             \
			     SPI_CONTEXT_CS_GPIO_SPEC_ELEM)

#define SPI_BITBANG_INIT(inst)                                               \
	static const struct spi_bitbang_config spi_bitbang_config_##inst = { \
		.regs = (void *)DT_INST_REG_ADDR(inst),                      \
		.clk = GPIO_DT_SPEC_INST_GET(inst, clk_gpios),               \
		.mosi = GPIO_DT_SPEC_INST_GET(inst, mosi_gpios),             \
		.cs = (const struct gpio_dt_spec[]){COND_CODE_1(             \
			DT_SPI_HAS_CS_GPIOS(inst),                           \
			(SPI_CONTEXT_CS_GPIOS_FOREACH_ELEM(inst)), ({0}))},  \
		.num_cs = DT_PROP_LEN_OR(inst, cs_gpios, 0),                 \
	};                                                                   \
                                                                             \
	static struct spi_bitbang_data spi_bitbang_data_##inst;              \
                                                                             \
	DEVICE_DT_INST_DEFINE(inst, spi_bitbang_init, NULL,                  \
			      &spi_bitbang_data_##inst,                      \
			      &spi_bitbang_config_##inst, POST_KERNEL,       \
			      CONFIG_SPI_INIT_PRIORITY, &spi_bitbang_api);

DT_INST_FOREACH_STATUS_OKAY(SPI_BITBANG_INIT)
